/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is NEON YUV-to-ARGB scaling code.
 *
 * The Initial Developer of the Original Code is ARM Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Jacob Bramley <Jacob.Bramley@arm.com>
 *   Steve McIntyre <Steve.McIntyre@linaro.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

    .arch   armv7-a
    .fpu    neon
    .text

    @ Provided by stdlib.h
    .global __udivsi3
    .type   __udivsi3, %function

    .global ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_asm
    .type   ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_asm, %function

@ ---- YUV Conversion Constants ----

    @ To decode Y, we must calculate (Y-16)*(255/219). Note that 255/219 rougly
    @ equals 74/64, which is easy to calculate with integer arithmetic.
#define     coeff_y_imm         74
    @ V->R: (255/112)*0.701*(V-128) = 1.596 * (V-128)
    @ In integer arithmetic, scale by 102/64.
#define     coeff_v_red_imm     102
    @ V->G: (255/112)*0.701*(0.299/0.587)*(V-128) = 0.813 * (V-128)
    @ In integer arithmetic, scale by 52/64.
#define     coeff_v_grn_imm     52
    @ U->G: (255/112)*0.886*(0.114/0.587)*(U-128) = 0.392 * (U-128)
    @ In integer arithmetic, scale by 25/64.
#define     coeff_u_grn_imm     25
    @ U->B: (255/112)*0.886*(U-128) = 2.017 * (U-128)
    @ In integer arithmetic, scale by 129/64.
#define     coeff_u_blu_imm     129

    @ Define bias values (for use as immediates).
#define     bias_y_imm          16
#define     bias_uv_imm         128

bias_table:
    @ Blue
    .short  bias_y_imm*coeff_y_imm + bias_uv_imm*coeff_u_blu_imm
    @ Green
    .short  bias_y_imm*coeff_y_imm - bias_uv_imm*(coeff_u_grn_imm+coeff_v_grn_imm)
    @ Red
    .short  bias_y_imm*coeff_y_imm + bias_uv_imm*coeff_v_red_imm
    @ Unused, but handy for debugging.
    .short  bias_y_imm*coeff_y_imm

    @ Define the common coefficient divisor using a power of two.
    @ 64 = 2^6
#define     coeff_divisor_imm   6

@ ---- Register Allocations ----

y               .req        r0      @ Source Y pointer.                          
u               .req        r1      @ Source U pointer.                          
v               .req        r2      @ Source V pointer.                          
dst             .req        r3      @ Destination row start pointer.             
dst_wh          .req        r4      @ Destination size (packed [16:16]).         
src_x           .req        r5      @ Source X co-ordinate (16.16 format).       
scr2            .req        r6      @ General scratch register.
src_dx          .req        r7      @ 16.16 format.                              
src_dy          .req        r8      @ 16.16 format.                              
args            .req        r9      @ Pointer to extra arguments.
dst_xy          .req        r10     @ Destination co-ordinates (packed [16:16]).  
scr0            .req        r11     @ General scratch register.                   
scr1            .req        r12     @ General scratch register.                  
@ Reserved                  r13     @ Stack pointer.                              
rstride_sd      .req        r14     @ Src and dst row-strides (packed [16:16]).   
@ Reserved                  r15     @ Program counter.                            

@ 8 D registers used for loading source data prior to filtering.
src_top_lo      .qn         q0
src_top_l_lo    .dn         d0      @ Top row of source pixels (lo).
src_top_r_lo    .dn         d1      @ Top row of source pixels (lo).
src_bot_lo      .qn         q1
src_bot_l_lo    .dn         d2      @ Bottom row of source pixels (lo).
src_bot_r_lo    .dn         d3      @ Bottom row of source pixels (lo).
src_top_hi      .qn         q2
src_top_l_hi    .dn         d4      @ Top row of source pixels (hi).    
src_top_r_hi    .dn         d5      @ Top row of source pixels (hi).
src_bot_hi      .qn         q3
src_bot_l_hi    .dn         d6      @ Bottom row of source pixels (hi).
src_bot_r_hi    .dn         d7      @ Bottom row of source pixels (hi). 

@ 4 Q registers used to hold ARGB data after colour-space conversion, but
@ before converting to RGB-565. These overlap with the source data registers
@ because the two set are never required at the same time.
blu             .qn         q0
blu_lo          .dn         d0
blu_hi          .dn         d1
grn             .qn         q1
grn_lo          .dn         d2
grn_hi          .dn         d3
red             .qn         q2
red_lo          .dn         d4
red_hi          .dn         d5
alp             .qn         q3
alp_lo          .dn         d6
alp_hi          .dn         d7

@ Filtered pixel data in the source (Y'CbCr) colour space.
y_filt          .qn         q4      @ 8 filtered Y' values.
y_filt_lo       .dn         d8
y_filt_hi       .dn         d9
u_filt          .qn         q5      @ 8 filtered Cb values.
u_filt_lo       .dn         d10
u_filt_hi       .dn         d11
v_filt          .qn         q6      @ 8 filtered Cr values.
v_filt_lo       .dn         d12
v_filt_hi       .dn         d13

@ Coefficients used for filtering.
w_row           .qn         q7      @ Row weightings (9.7 format).

w_col           .qn         q8      @ Column weightings (1.7 format).
w_col_lo        .dn         d16
w_col_hi        .dn         d17

@ Coefficients used for colour-space conversion.
contr_y_lo      .qn         q8      @ Overlaps with w_col.
contr_y_hi      .qn         q11     @ Overlaps with qscr0 and coefficients.

bias_packed     .dn         d18
bias_r          .qn         q1      @ Overlaps with grn.
bias_g          .qn         q2      @ Overlaps with red.
bias_b          .qn         q3      @ Overlaps with alp.

coeff_y         .dn         d19
coeff_u_blu     .dn         d20
coeff_v_red     .dn         d21
coeff_v_grn     .dn         d0      @ Overlaps with blu.
coeff_u_grn     .dn         d1      @ Overlaps with blu.

@ Scratch registers.
qscr0           .qn         q11     @ Scratch register.
qscr0_lo        .dn         d22
qscr0_hi        .dn         d23

qscr1           .qn         q12     @ Scratch register.
qscr1_lo        .dn         d24
qscr1_hi        .dn         d25

qscr2           .qn         q13     @ Scratch register.
qscr2_lo        .dn         d26
qscr2_hi        .dn         d27

qscr3           .qn         q14     @ Scratch register.
qscr3_lo        .dn         d28
qscr3_hi        .dn         d29

qscr4           .qn         q15     @ Scratch register.
qscr4_lo        .dn         d30
qscr4_hi        .dn         d31

@ void LinearScaleYV12ToRGB32_NEON(
@   uint8_t const *  y,             // Buffer of source 'y' values.
@   uint8_t const *  u,             // Buffer of source 'u' values.
@   uint8_t const *  v,             // Buffer of source 'v' values.
@   uint8_t       *  dst,           // Buffer for destination ARGB-8888 values.
@   int              src_width,     // Source width.
@   int              src_height,    // Source height.
@   int              dst_width,     // Destination width.
@   int              dst_height,    // Destination height.
@   int              src_rowstride, // Source row-stride.
@   int              dst_rowstride) // Destination row-stride.

    .balign 64
    .fnstart
ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_asm:

    @ Temporary allocations. These values are either not required for long, or
    @ are packed to save register space.
    src_w       .req        r4
    src_h       .req        r5
    dst_w       .req        r6
    dst_h       .req        r9
    src_rstride .req        r10
    dst_rstride .req        r11

    mov     args,   sp

    push    {r4-r12, lr}    @ 10 words.
    vpush   {q4-q7}         @ 16 words.
    
    @ Load extra arguments.
    ldm     args,   {src_w,src_h,dst_w,dst_h,src_rstride,dst_rstride}

    @ The filter works by considering each source pixel to be a point.
    @ Destination pixels are populated by a weighted average of the
    @ surrounding four points. This filter should provide good results for
    @ photographic sources, and sources with few hard edges (such as most
    @ videos). For sources with hard edges, consider using an alternative
    @ filter.
    @
    @ Note: Scaling factors smaller than 0.5 may benefit from sampling more
    @ than four source pixels for each destination pixel. However, this
    @ requires a separate filter implementation, and such small scales are
    @ unlikely to be useful for internet videos.
    @
    @   • For each dst row:
    @       • Work out which two source rows will contribute.
    @         - The source rows are not always adjacent. For the top and bottom
    @           rows, the source rows will be identical.
    @       • Work out the fractional contribution from each source row.
    @
    @       • For each dst column:
    @           • Work out which two source columns will contribute.
    @             - The source columns are not always adjacent.
    @           • Work out the fractional contribution from each source column.
    @           • Calculate a weighted average of each source column pair.
    @
    @       • Calculate a weighted average of each source row pair.
    @

    @ ==== Scale: 0.5 to 2.0 ====
    @ Scaling to larger destinations is possible, but image quality will be
    @ poor as the Cr and Cb channels are sampled using nearest-pixel filtering.

    @ TODO: Check the constraints of this scaler.
    @  Max source width: 65535
    @  Max source height: 65535
    @  Min destination width: 8
    @  Src row-stride: 65535
    @  Dst row-stride: 65535
    
    @ Calculate the source stepping in each direction.

    push    {r0-r3,ip,lr}

    @ When upscaling, we find src_dx and src_dy as follows:
    @  src_dx = (src_w-1)/(dst_w-1)
    @  src_dy = (src_h-1)/(dst_h-1)
    @ This ensures that the sample points lie within the source image.
    @
    @ When downscaling, this doesn't work, but instead we can use src_w/dst_w.

    cmp     src_h, dst_h
    mov     r0, src_h, LSL #16  @ Convert to 16.16 format.
    itte    lo
    sublo   r0, #(1<<16)
    sublo   r1, dst_h, #1
    movhs   r1, dst_h
    bl      __udivsi3
    mov     src_dy, r0

    cmp     src_w, dst_w
    mov     r0, src_w, LSL #16  @ Convert to 16.16 format.
    itte    lo
    sublo   r0, #(1<<16)
    sublo   r1, dst_w, #1
    movhs   r1, dst_w
    bl      __udivsi3
    mov     src_dx, r0

    pop     {r0-r3,r12,lr}

    @ Pack dst_w and dst_h into one 32-bit register.
    bfi     dst_wh, dst_w, #16, #16 @ dst_wh[31:16] = dst_w[15: 0]
    bfi     dst_wh, dst_h, # 0, #16 @ dst_wh[15: 0] = dst_h[15: 0]

    @ Pack src_rstride and dst_rstride into one 32-bit register.
    bfi     rstride_sd, src_rstride, #16, #16
    bfi     rstride_sd, dst_rstride, # 0, #16

    @ Stack some values for retrieval later.
    push    {y,u,v,dst}

    @ Initialize YUV->RGB conversion constants. We don't have enough registers
    @ to initialize them all at this stage, so some are constructed in the loop.
    adr         scr0, bias_table
    vmov.u8     coeff_y, #coeff_y_imm
    vmov.u8     coeff_u_blu, #coeff_u_blu_imm
    vmov.u8     coeff_v_red, #coeff_v_red_imm
    vld1.u16    {bias_packed}, [scr0]

    @ Start in the top left corner. When down-scaling, start slightly offset so
    @ we don't take samples exactly from every source, particularly when
    @ scaling by 1/2, 1/3, 1/4 and so on.
    @ Use an offset of (src_dy-1)/2, clamped to 0.
    subs    scr0, src_dy, #(1<<16)  @ 16.16 format.
    movmi   scr0, #0
    movpl   scr0, scr0, LSL #15     @ Divide by two and convert to 0.32.
    mov     dst_xy, #0

1:  @ ==== Row loop. ====
    @ At this point, scr0 should contain the fractional part of the src Y
    @ co-ordinate.

    @ Calculate the contribution from each row.
    @  - The contribution from each row is equal to one minus the distance from
    @    the row, where the distance between two rows is 1. The fractional part
    @    of src_y holds the distance from the top row, so we can derive both
    @    rows from that.
    @  - The range of the top row contribution is:      0 >  t >= 1.0
    @  - The range of the bottom row contribution is:   0 >= b >  1.0
    ubfx        scr0, scr0, #25, #7
    vdup.u16    w_row, scr0
    @ w_row.u8[] = bottom row contributions (0.7 format).

    @ Initialize src_x.
    @ This is fractional in the same way as the initial value of src_y.
    subs    src_x, src_dx, #(1<<16)
    movmi   src_x, #0
    movpl   src_x, src_x, LSR #1

2:  @ ==== Column loop. ====
    @ Load source values and calculate bilinear filtering co-efficients.
    @ Unpack src_rstride.
    ubfx    scr1, rstride_sd, #16, #16
    .irp    dword,lo,hi
        .irp    lane,0,1,2,3,4,5,6,7
            add     scr0, y, src_x, LSR #16
            @ - Load two Y elements from each row.
            vld2.u8 {src_top_l_\dword[\lane], src_top_r_\dword[\lane]}, [scr0], scr1
            vld2.u8 {src_bot_l_\dword[\lane], src_bot_r_\dword[\lane]}, [scr0]
            @ - Load the nearest U and V elements.
            add     scr0, u, src_x, LSR #17
            add     scr2, v, src_x, LSR #17
            vld1.u8 {u_filt_\dword[\lane]}, [scr0]
            vld1.u8 {v_filt_\dword[\lane]}, [scr2]
            @ - Calculate the (right side) weightings, as for the row weightings.
            ubfx    scr0, src_x, #9, #7
            vmov.u8 w_col_\dword[\lane], scr0   @ Fixed-point 1.7 format.
            @ - Advance to the next source column.
            add     src_x, src_dx
        .endr
    .endr

    @ Calculate the contribution from each column.
    @  - Get the value "1" in 1.7 format.
    vmov.u8     qscr4,  #(1<<7)
    @  - Calculate right-side contributions.
    vmull.u8    qscr0,  src_bot_r_lo,   w_col_lo
    vmull.u8    qscr1,  src_bot_r_hi,   w_col_hi
    vmull.u8    qscr2,  src_top_r_lo,   w_col_lo
    vmull.u8    qscr3,  src_top_r_hi,   w_col_hi
    @  - Calculate 1-w_col to get left-side contributions.
    vsub.u8     w_col, qscr4, w_col
    @  - Accumulate left-side contributions.
    vmlal.u8    qscr0,  src_bot_l_lo,   w_col_lo
    vmlal.u8    qscr1,  src_bot_l_hi,   w_col_hi
    vmlal.u8    qscr2,  src_top_l_lo,   w_col_lo
    vmlal.u8    qscr3,  src_top_l_hi,   w_col_hi
    @  - Shift down to the 0-255 range.
    vrshr.u16   qscr0,  qscr0,  #7
    vrshr.u16   qscr1,  qscr1,  #7
    vrshr.u16   qscr2,  qscr2,  #7
    vrshr.u16   qscr3,  qscr3,  #7
    @ qscr0 = bottom row (lo)
    @ qscr1 = bottom row (hi)
    @ qscr2 = top row (lo)
    @ qscr3 = top row (hi)
    @  - Get the value "1" in 1.15 format.
    vmov.u16    qscr4,  #(1<<7)
    @  - Calculate bottom-row contributions.
    vmul.u16    qscr0,  qscr0,  w_row
    vmul.u16    qscr1,  qscr1,  w_row
    @  - Invert w_row to get top-side contributions.
    vsub.u16    w_row, qscr4, w_row
    @  - Accumulate top-row contributions.
    vmla.u16    qscr0,  qscr2,  w_row
    vmla.u16    qscr1,  qscr3,  w_row
    @  - Shift down by 7 bits and narrow into the filtered output register.
    vrshrn.u16  y_filt_lo,  qscr0,  #7
    vrshrn.u16  y_filt_hi,  qscr1,  #7
    @  - Reset w_row.
    vsub.u16    w_row, qscr4, w_row

    @ Convert y_filt, u_filt and v_filt to RGB.
    @  - Calculate the common Y contribution for each channel.
    vmull.u8    contr_y_lo, y_filt_lo, coeff_y
    vmull.u8    contr_y_hi, y_filt_hi, coeff_y
    @  - Initialize conversion co-efficients that weren't already initialized.
    vmov.u8     coeff_v_grn, #coeff_v_grn_imm
    vmov.u8     coeff_u_grn, #coeff_u_grn_imm
    @  - Initialize staging areas for R and B. G can use y_contr.
    vmov        qscr1, contr_y_lo   @ Staging area for B.
    vmov        qscr2, contr_y_hi
    vmov        qscr3, contr_y_lo   @ Staging area for R.
    vmov        qscr4, contr_y_hi
    vdup.s16    bias_b, bias_packed[0]  @ Unpack B bias.
    @  - Negative U contributions for G.
    vmlsl.u8    contr_y_lo, u_filt_lo, coeff_u_grn
    vmlsl.u8    contr_y_hi, u_filt_hi, coeff_u_grn
    vdup.s16    bias_g, bias_packed[1]  @ Unpack G bias.
    @  - Positive U contributions for B.
    vmlal.u8    qscr1, u_filt_lo, coeff_u_blu
    vmlal.u8    qscr2, u_filt_hi, coeff_u_blu
    vdup.s16    bias_r, bias_packed[2]  @ Unpack R bias.
    @  - Positive V contributions for R.
    vmlal.u8    qscr3, v_filt_lo, coeff_v_red
    vmlal.u8    qscr4, v_filt_hi, coeff_v_red
    @  - Negative V contributions for G.
    vmlsl.u8    contr_y_lo, v_filt_lo, coeff_v_grn
    vmlsl.u8    contr_y_hi, v_filt_hi, coeff_v_grn
    @  - Add bias.
    vsub.s16    qscr1, bias_b
    vsub.s16    qscr2, bias_b
    vsub.s16    qscr3, bias_r
    vsub.s16    qscr4, bias_r
    vsub.s16    contr_y_lo, bias_g
    vsub.s16    contr_y_hi, bias_g

    @  - Scale back to 8 bits.
    vqrshrun.s16    blu_lo, qscr1, #coeff_divisor_imm
    vqrshrun.s16    blu_hi, qscr2, #coeff_divisor_imm
    vqrshrun.s16    red_lo, qscr3, #coeff_divisor_imm
    vqrshrun.s16    red_hi, qscr4, #coeff_divisor_imm
    vqrshrun.s16    grn_lo, contr_y_lo, #coeff_divisor_imm
    vqrshrun.s16    grn_hi, contr_y_hi, #coeff_divisor_imm

    @  - Convert to RGB-565 (lo).
    vshll.u8        qscr0,  red_lo, #8
    vshll.u8        qscr2,  grn_lo, #8
    vshll.u8        qscr3,  blu_lo, #8
    vsri.u16        qscr0,  qscr2,  #5
    vsri.u16        qscr0,  qscr3,  #11

    @  - Convert to RGB-565 (hi).
    vshll.u8        qscr1,  red_hi, #8
    vshll.u8        qscr2,  grn_hi, #8
    vshll.u8        qscr3,  blu_hi, #8
    vsri.u16        qscr1,  qscr2,  #5
    vsri.u16        qscr1,  qscr3,  #11

    @ Store the ARGB result.
    vst1.u16 {qscr0_lo,qscr0_hi,qscr1_lo,qscr1_hi}, [dst]!

    @ End the column loop.
    @ There are three conditions to test:
    @ Let d = dst_w - dst_x.
    @ d >= 32:          The next block fits wholly within the destination image.
    @ d < 32 && d > 16: The next block fits partially within the destination image.
    @ d <= 16:          The next block does not fit in the destination image.
    
    @ Common case: The next block fits wholly within the destination image.
    mov     scr0, dst_wh, LSR #16
    sub     scr0, dst_xy, LSR #16   @ Find 'd'.
    cmp     scr0, #32
    @ Advance dst_xy (in 16:16 format) to the next block.
    add     dst_xy, #(16<<16)

    bhs     2b  @ Loop if d >= 23.
    
    cmp     scr0, #16
    ble     3f  @ Exit column loop if d <= 16.

    @ The next block overlaps the destination image border. We apply an offset
    @ to dst_x to make the block fit.
    rsb     scr0,   #32
    mul     scr1,   scr0, src_dx
    sub     dst,    scr0, LSL #1    @ 2 bytes per pixel.
    sub     dst_xy, scr0, LSL #16
    sub     src_x,  scr1
    b       2b  @ Loop.

3:  @ ==== End of column loop. ====
    
    @ End the row loop.

    @ Reset the source and destination pointers.
    ldm     sp, {y,u,v,dst}

    @ Advance dst_xy to the next row.
    bfc     dst_xy, #16, #16
    add     dst_xy, #1

    @ Check that we're still within the destination image.
    @ Note that dst_y has already been incremented for the next row.
    @ Let d = dst_h - dst_y
    @ d >  0: The next destination row is within the destination image.
    @ d <= 0: The next destination row is outside the destination image.
    mov     scr1, dst_wh, LSL #16
    cmp     scr1, dst_xy, LSL #16
    ble     4f  @ Exit the column loop if d <= 0.

    @ Calculate src_y.
    mov     scr0, dst_xy, LSL #16       @ Get dst_y in 16.16 format.
    umull   scr0, scr1, src_dy, scr0    @ Result is in 32.32 format.
    @   scr0 = src_y, fractional part.
    @   scr1 = src_y, integral part.
    @ If we're down-scaling (and thus src_dy >= 1), add the initial offset of
    @ (src_dy-1)/2 to src_y.
    subs    scr2, src_dy, #(1<<16)
    movmi   scr2, #0
    adds    scr0, scr2, LSL #15 @ Divide by 2 and convert to 0.32, then add.
    adc     scr1, scr2, LSR #17 @ Divide by 2 and convert to 32.0, then add.

    @ y += src_y * src_rstride
    ubfx    scr2, rstride_sd, #16, #16
    mla     y, scr1, scr2, y
    @ u += (src_y/2 * src_rstride/2)
    mov     scr1, scr1, LSR #1  @ Find src_y/2.
    mov     scr2, scr2, LSR #1  @ Find src_rstride/2.
    mla     u, scr2, scr1, u
    @ u += (src_y/2 * src_rstride/2)
    mla     v, scr2, scr1, v
    @ dst += (dst_y * dst_rstride)
    ubfx    scr1, dst_xy, #0, #16       @ Find dst_y.
    mov     scr2, rstride_sd, LSL #1    @ Find dst_rstride (* 2 bytes per pixel).
    bfc     scr2, #16, #15
    mla     dst, scr1, scr2, dst

    @ At this point, scr0 must contain the fractional part of scr_y, because it
    @ is required at the start of the row loop.

    @ If ((dst_h-dst_y)*src_dy) < 0:
    @   The next two source rows are partially within the source image.
    @ else:
    @   The next two source rows are wholly within the source image.
    mov     scr1, dst_wh, LSL #16
    sub     scr1, dst_xy, LSL #16
    mov     scr1, scr1, ASR #16
    mul     scr1, src_dy, scr1      @ scr = (dst_h-dst_y)*src_dy, 16.16 format.
    cmp     scr1, #0
    bge     1b  @ The next two source rows are wholly within the source image.

    bfc     rstride_sd, #16, #16    @ Make src_rstride 0 to prevent over-read.
    stm     sp, {y,u,v}
    b       1b
    
4:  @ ==== End of row loop. ====

9:  @ We're either finished, or we've given up.
    
    add     sp, sp, #(4*4)

    vpop    {q4-q7}
    pop     {r4-r12, pc}

    .fnend
