/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is NEON YUV-to-ARGB scaling code.
 *
 * The Initial Developer of the Original Code is ARM Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Jacob Bramley <Jacob.Bramley@arm.com>
 *   Steve McIntyre <Steve.McIntyre@linaro.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

    .arch   armv7-a
    .fpu    neon
    .text

    @ Provided by stdlib.h
    .global __udivsi3
    .type   __udivsi3, %function

    .global ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_vtbl_asm
    .type   ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_vtbl_asm, %function

@ ---- YUV Conversion Constants ----

    @ To decode Y, we must calculate (Y-16)*(255/219). Note that 255/219 rougly
    @ equals 74/64, which is easy to calculate with integer arithmetic.
#define     coeff_y_imm         74
    @ V->R: (255/112)*0.701*(V-128) = 1.596 * (V-128)
    @ In integer arithmetic, scale by 102/64.
#define     coeff_v_red_imm     102
    @ V->G: (255/112)*0.701*(0.299/0.587)*(V-128) = 0.813 * (V-128)
    @ In integer arithmetic, scale by 52/64.
#define     coeff_v_grn_imm     52
    @ U->G: (255/112)*0.886*(0.114/0.587)*(U-128) = 0.392 * (U-128)
    @ In integer arithmetic, scale by 25/64.
#define     coeff_u_grn_imm     25
    @ U->B: (255/112)*0.886*(U-128) = 2.017 * (U-128)
    @ In integer arithmetic, scale by 129/64.
#define     coeff_u_blu_imm     129

    @ Define bias values.
#define     bias_y_imm          16
#define     bias_u_imm          128
#define     bias_v_imm          128

    @ Define the coefficient divisor (common to all channels) using a power of
    @ two: 64 = 2^6
#define     coeff_divisor_imm   6

    .balign 4
bias_table:     @ .s16
    @ [0]: Blue
    .short  bias_y_imm*coeff_y_imm + bias_u_imm*coeff_u_blu_imm
    @ [1]: Green
    .short  bias_y_imm*coeff_y_imm - bias_u_imm*coeff_u_grn_imm - bias_v_imm*coeff_v_grn_imm
    @ [2]: Red
    .short  bias_y_imm*coeff_y_imm + bias_v_imm*coeff_v_red_imm
    @ [3]: Unused, but handy for debugging with U and V disabled.
    .short  bias_y_imm*coeff_y_imm

    .balign 4
coeff_table:    @ .u8
    .byte   coeff_y_imm     @ [0]
    .byte   coeff_v_red_imm @ [1]
    .byte   coeff_v_grn_imm @ [2]
    .byte   coeff_u_grn_imm @ [3]
    .byte   coeff_u_blu_imm @ [4]

@ ---- Register Allocations ----

@ -- General-purpose registers. --

y               .req        r0      @ Source Y pointer.                          
u               .req        r1      @ Source U pointer.                          
v               .req        r2      @ Source V pointer.                          
dst             .req        r3      @ Destination row start pointer.             
dst_wh          .req        r4      @ Destination size (packed [16:16]).         
src_x           .req        r5      @ Source X co-ordinate (16.16 format).       
scr2            .req        r6      @ General scratch register.
src_dx          .req        r7      @ 16.16 format.                              
src_dy          .req        r8      @ 16.16 format.                              
@ w_row_packed is used in the scalar variant of VMUL and VMLA, and as such is restricted
scr3            .req        r9      @ General scratch register.                  
dst_xy          .req        r10     @ Destination co-ordinates (packed [16:16]).  
scr0            .req        r11     @ General scratch register.                   
scr1            .req        r12     @ General scratch register.                  
@ Unavailable               r13     @ Stack pointer.                              
rstride_sd      .req        r14     @ Src and dst row-strides (packed [16:16]).   
@ Unavailable               r15     @ Program counter.                            

@ We don't have enough NEON registers to allocate a unique register for every
@ quantity that we need to keep track of. To facilitate efficient register
@ re-use whilst also trying to reduce risk of accidental (invalid) overlapping,
@ the algorithm is divided into phases:
@
@  * Row Loop
@    * Row Set-Up
@    * Column Loop
@      * Column Set-Up
@      * Load
@      * Filter
@      * RGB-565
@      * Store
@      * Column Iteration
@    * Row Iteration
@
@ Registers are reserved for contiguous phases (or around the end of the loop).
@
@ Registers may still overlap within phases, but these overlaps are localized
@ and relatively easy to manage.

@ -- Reserved throughout. --
bias_packed     .dn         d0  @ See bias_table for element packing.
coeff_packed    .dn         d1  @ See coeff_table for element packing.

@ w_row_packed is used in the scalar variant of VMUL and VMLA, and as such is
@ restricted to d0-d7.
w_row_packed    .dn         d2  @ [0]: bottom row weight. [1]: top row weight.

@ -- Scratch for "Row Set-Up". --
src_x0v         .qn     q15

@ -- Reserved from "Row Set-Up" onwards. --
src_xv_0        .qn     q2
src_xv_0_lo     .dn         d4
src_xv_0_hi     .dn         d5
src_xv_1        .qn     q3
src_xv_1_lo     .dn         d6
src_xv_1_hi     .dn         d7
src_xv_2        .qn     q4
src_xv_2_lo     .dn         d8
src_xv_2_hi     .dn         d9
src_xv_3        .qn     q5
src_xv_3_lo     .dn         d10
src_xv_3_hi     .dn         d11

@ -- Scratch for "Column Set-Up". --

sample_0        .qn     q6
sample_0_lo     .dn         d12
sample_0_hi     .dn         d13
sample_1        .qn     q7
sample_1_lo     .dn         d14
sample_1_hi     .dn         d15

step            .qn     q12

offset          .qn     q12
offset_uv       .qn     q13

@ -- Reserved from "Column Set-Up" to "Load". --
index           .qn     q8
index_lo        .dn         d16
index_hi        .dn         d17

index_uv        .qn     q6          @ Overwrites sample_0.
index_uv_lo     .dn         d12
index_uv_hi     .dn         d13

@ -- Reserved from "Column Set-Up" to "Filter". --
w_col_right     .qn     q9
w_col_right_lo  .dn         d18
w_col_right_hi  .dn         d19

@ -- Scratch for "Load". --
index2          .qn     q7
index2_lo       .dn         d14
index2_hi       .dn         d15

data0           .dn         d20
data1           .dn         d21
data2           .dn         d22
data3           .dn         d23

@ -- Reserved from "Load" until "Filter". --
src_top_l       .qn     q12
src_top_l_lo    .dn         d24
src_top_l_hi    .dn         d25
src_top_r       .qn     q13
src_top_r_lo    .dn         d26
src_top_r_hi    .dn         d27
src_bot_l       .qn     q14
src_bot_l_lo    .dn         d28
src_bot_l_hi    .dn         d29
src_bot_r       .qn     q15
src_bot_r_lo    .dn         d30
src_bot_r_hi    .dn         d31

@ -- Reserved from "Load" until "RGB-565". --
u_filt          .qn     q7          @ 8 filtered Cb values. Overwrites index2.
u_filt_lo       .dn         d14
u_filt_hi       .dn         d15 
v_filt          .qn     q6          @ 8 filtered Cr values. Overwrites index_uv.
v_filt_lo       .dn         d12
v_filt_hi       .dn         d13

@ -- Scratch for "Filter". --
w_col_left      .qn     q8
w_col_left_lo   .dn         d16
w_col_left_hi   .dn         d17

w_row_bot       .qn     q8          @ Overwrites w_col_left.
w_row_top       .qn     q9          @ Overwrites w_col_right.

filt_bot_lo     .qn     q10
filt_bot_hi     .qn     q11
filt_top_lo     .qn     q15         @ Overwrites src_bot_r.
filt_top_hi     .qn     q13         @ Overwrites src_top_r.

filt_lo         .qn     q10         @ Overwrites filt_bot_lo.
filt_hi         .qn     q11         @ Overwrites filt_bot_hi.

@ -- Reserved from "Filter" until "RGB-565". --
y_filt          .qn     q15         @ Overwrites filt_top_lo.
y_filt_lo       .dn         d30
y_filt_hi       .dn         d31

@ -- Scratch for "RGB-565". --
blu16_lo        .qn     q8
blu16_lo_lo     .dn         d16
blu16_lo_hi     .dn         d17
blu16_hi        .qn     q9
blu16_hi_lo     .dn         d18
blu16_hi_hi     .dn         d19
grn16_lo        .qn     q10
grn16_lo_lo     .dn         d20
grn16_lo_hi     .dn         d21
grn16_hi        .qn     q11
grn16_hi_lo     .dn         d22
grn16_hi_hi     .dn         d23
red16_lo        .qn     q12
red16_lo_lo     .dn         d24
red16_lo_hi     .dn         d25
red16_hi        .qn     q13
red16_hi_lo     .dn         d26
red16_hi_hi     .dn         d27

coeff_y         .dn     d3
coeff_u_grn     .dn     d28
coeff_u_blu     .dn     d29
coeff_v_grn     .dn     d30         @ Overwrites y_filt.
coeff_v_red     .dn     d31         @ Overwrites y_filt.

bias_r          .qn     q14         @ Overwrites coeff_u_grn and coeff_u_blu.
bias_g          .qn     q7          @ Overwrites u_filt.
bias_b          .qn     q15         @ Overwrites y_filt.

red             .qn     q6          @ Overwrites bias_r.
red_lo          .dn         d12
red_hi          .dn         d13
grn             .qn     q7          @ Overwrites bias_g.
grn_lo          .dn         d14
grn_hi          .dn         d15
blu             .qn     q15         @ Overwrites bias_b.
blu_lo          .dn         d30   
blu_hi          .dn         d31

@ Reserved from "RGB-565" until "Store".
rgb565_lo       .qn     q12         @ Overwrites red16_lo.
rgb565_lo_lo    .dn         d24
rgb565_lo_hi    .dn         d25
rgb565_hi       .qn     q13         @ Overwrites red16_hi.
rgb565_hi_lo    .dn         d26
rgb565_hi_hi    .dn         d27

@ Scratch for "Column Iteration" and "Row Iteration".
src_xv_step     .qn     q15

@ void LinearScaleYV12ToRGB32_NEON(
@   uint8_t const *  y,             // Buffer of source Y' values.
@   uint8_t const *  u,             // Buffer of source Cb values.
@   uint8_t const *  v,             // Buffer of source Cr values.
@   uint8_t       *  dst,           // Buffer for destination ARGB-8888 values.
@   int              src_width,     // Source width.
@   int              src_height,    // Source height.
@   int              dst_width,     // Destination width.
@   int              dst_height,    // Destination height.
@   int              src_rowstride, // Source row-stride.
@   int              dst_rowstride) // Destination row-stride.

@ Hard constraints:
@  Maximum source width: 65535
@  Maximum source height: 65535
@  Minimum source height: 2
@  Minimum destination width: 16
@  Minimum scaling factor: 0.5
@  Maximum scaling factor: Very high *.
@  Maximum source row-stride: 65535 pixels
@  Maximum destination row-stride: 65535 pixels
@
@ * Whilst the filter does produce output for very large scaling factors (such
@   as 16), the Cb and Cr channels are sampled using nearest-pixel filtering,
@   and so the output starts to look blocky at factors of about 1.6. Some
@   sources scale better than others.
@
@ The filter may over-read the source rows by 16 pixels. The source must have
@ enough padding to allow this.
@
@ This filter can be applied to IYUV as well as YV12. For IYUV, swap the 'u' and
@ 'v' inputs.
@
@ The output buffer must be 2-byte-aligned. There are no alignment constraints
@ on the input buffers.
@
@ TODO: Are there any other constraints?

    .balign 64
    .fnstart
ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_vtbl_asm:

    @ Temporary allocations. These values are either not required for long, or
    @ are subsequently packed to save register space.
    src_w       .req        r4
    src_h       .req        r5
    dst_w       .req        r6
    dst_h       .req        r9
    src_rstride .req        r10
    dst_rstride .req        r11

    mov     scr3,   sp

    @ Deal with callee-saved registers. We need to use every register we have.
    push    {r4-r12, lr}
    vpush   {q4-q7}
    
    @ Load extra arguments from the stack.
    ldm     scr3,   {src_w,src_h,dst_w,dst_h,src_rstride,dst_rstride}

    @ The filter works by considering each source pixel to be a point.
    @ Destination pixels are populated by a weighted average of the
    @ surrounding four points. This filter should provide good results for
    @ photographic sources, and sources with few hard edges (such as most
    @ videos). For images with hard edges, a more suitable scaler should be
    @ used.
    @
    @ Also note that border pixels are slightly under-represented in the output
    @ when up-scaling. This is done to avoid having to deal with negative
    @ co-ordinates, or co-ordinates outside the source image. The result should
    @ be acceptable for typical video sources, however.

    @ ==== Scale, bilinear Y: 0.5 and above. ====
    @ Scaling to larger destinations is possible, but image quality will be
    @ poor as the Cr and Cb channels are sampled using nearest-pixel filtering.

    @ Calculate the source stepping in each direction.
    @ When upscaling, we find src_dx and src_dy as follows:
    @  src_dx = (src_w-1)/(dst_w-1)
    @  src_dy = (src_h-1)/(dst_h-1)
    @ This ensures that the sample points lie within the source image.
    @ When downscaling, this doesn't work, but instead we can use src_w/dst_w.
    @
    @ Because this code is only run once per frame, we just fall back to the
    @ platform library's division function (for simplicity).

    push    {r0-r3,ip,lr}       @ Preserve caller-saved registers.

    cmp     src_h, dst_h
    mov     r0, src_h, LSL #16  @ Convert to 16.16 format.
    itte    lo
    sublo   r0, #(1<<16)
    sublo   r1, dst_h, #1
    movhs   r1, dst_h
    bl      __udivsi3
    mov     src_dy, r0

    cmp     src_w, dst_w
    mov     r0, src_w, LSL #16  @ Convert to 16.16 format.
    itte    lo
    sublo   r0, #(1<<16)
    sublo   r1, dst_w, #1
    movhs   r1, dst_w
    bl      __udivsi3
    mov     src_dx, r0

    pop     {r0-r3,r12,lr}      @ Restore caller-saved registers.

    @ Pack dst_w and dst_h into one 32-bit register.
    bfi     dst_wh, dst_w, #16, #16 @ dst_wh[31:16] = dst_w[15: 0]
    bfi     dst_wh, dst_h, # 0, #16 @ dst_wh[15: 0] = dst_h[15: 0]

    @ Pack src_rstride and dst_rstride into one 32-bit register.
    bfi     rstride_sd, src_rstride, #16, #16
    bfi     rstride_sd, dst_rstride, # 0, #16

    @ Stack some values for retrieval later.
    push    {y,u,v,dst}

    @ Initialize YUV->RGB conversion constants. We don't have enough registers
    @ to have unique registers throughout the conversion loop, so these are
    @ packed into D registers, to be unpacked using VDUP instructions
    @ when required.
    adr         scr0, bias_table
    adr         scr1, coeff_table
    vld1.s16    {bias_packed}, [scr0]
    vld1.u8     {coeff_packed}, [scr1]

    @ Start in the top left corner. When down-scaling, start slightly offset.
    @ Compute src_y, but don't store it anywhere. It can be quickly re-computed
    @ at the end of each row iteration.
    @ Start with an offset of (src_dy-1)/2, saturating at 0.
    subs        scr0, src_dy, #(1<<16)  @ (src_dy-1.0)      {16.16}
    movmi       scr0, #0
    @ Find frac((src_dy-1)/2) and convert to {0.7} in one step.
    ubfx        scr0, scr0, #9, #7
    @ Initialize the destination co-ordinates.
    mov         dst_xy, #0

    @ Calculate the contribution from each row.
    @  - The contribution from each row is equal to one minus the distance from
    @    the row, where the distance between two rows is 1. The fractional part
    @    of src_y holds the distance from the top row, so we can derive both
    @    rows from that.
    @  - The range of the top row contribution is:      0 >  t >= 1.0
    @  - The range of the bottom row contribution is:   0 >= b >  1.0
    rsb         scr1, scr0, #0x0080     @ 1.0 in 9.7 format.
    vmov.u16    w_row_packed[0], scr0
    vmov.u16    w_row_packed[1], scr1

1:  @ ==== Row loop. ====

    @ Phase: Row Set-Up

    @ Initialize src_x.
    @ This is fractional in the same way as the initial value of src_y.
    subs        src_x, src_dx, #(1<<16)
    movmi       src_x, #0
    mov         src_x, src_x, LSR #1

    @ Initialize the src_xv vectors for the start of the row (dst_x=0).
    mov         scr1, src_dx, LSL #1        @ src_dx * 2
    mov         scr2, src_dx, LSL #2        @ src_dx * 4
    mov         scr3, src_dx, LSL #3        @ src_dx * 8
    vdup.u32    src_x0v, src_x
    vmov.u32    src_xv_0_lo, #0
    vmov.u32    src_xv_0_lo[1], src_dx      @ < 0| 0| 1| 0> * src_dx
    vdup.u32    src_xv_0_hi, scr1           @ < 2| 2| 1| 0> * src_dx
    vadd.u32    src_xv_0_hi, src_xv_0_lo    @ < 3| 2| 1| 0> * src_dx
    vadd.u32    src_xv_0, src_x0v           @ < 3| 2| 1| 0> * src_dx + src_x0
    add         scr1, scr2, scr3
    vdup.u32    src_xv_1, scr2              @ < 4| 4| 4| 4>
    vdup.u32    src_xv_2, scr3              @ < 8| 8| 8| 8>
    vdup.u32    src_xv_3, scr1              @ <12|12|12|12>
    vadd.u32    src_xv_1, src_xv_0          @ < 7| 6| 5| 3> * src_dx + src_x0
    vadd.u32    src_xv_2, src_xv_0          @ <11|10| 9| 8> * src_dx + src_x0
    vadd.u32    src_xv_3, src_xv_0          @ <15|14|13|12> * src_dx + src_x0

2:  @ ==== Column loop. ====

    @ Phase: Column Set-Up

    @ Work out the index and column weighting vectors for this block.
    @
    @ Each src_x (Y') value is related to dst_x as follows:
    @  src_x[x] = src_x0 + dst_x * src_dx
    @ For Cb/U and Cr/V:
    @  src_x_uv[x] = floor(round(src_x[x])/2)
    @ 
    @ In the load stage, we load blocks of 16 source pixels and use VTBL to
    @ pick out the relevant pixels. First, we need to construct the index
    @ vectors and find the base offset for each block.

    mov         scr0, src_dx, LSL #4        @ src_dx * 16
    
    @ For each lane of src_xv:
    @ Bits [23:16] of each element form the index.
    @ Bits [15: 9] of each element form the column (right-side) weighting, for
    @ linear scaling.
    @ We can throw the other bits away as they are not relevant here.
    vshrn.u32   sample_0_lo, src_xv_0, #8   @ Sample block 0. {8.8}
    vshrn.u32   sample_0_hi, src_xv_1, #8   @ Sample block 1. {8.8}
    vshrn.u32   sample_1_lo, src_xv_2, #8   @ Sample block 2. {8.8}
    vshrn.u32   sample_1_hi, src_xv_3, #8   @ Sample block 3. {8.8}
    @ Initialize 'step', to advance src_xv_[0-3] for the next iteration.
    vdup.u32    step, scr0
    @ Extract the index vectors and column weights.
    vmovn.u16   w_col_right_lo, sample_0
    vshrn.u16   index_lo, sample_0, #8
    vmovn.u16   w_col_right_hi, sample_1
    vshrn.u16   index_hi, sample_1, #8
    @ Convert column weights to 1.7 format.
    vshr.u8     w_col_right, #1
    @ Calculate the index vector for Cb/U and Cr/V.
    @  src_x_uv[x] = floor(round(src_x[x])/2)
    vrshrn.u16  index_uv_lo, sample_0, #8   @ round(src_x)
    vrshrn.u16  index_uv_hi, sample_1, #8   @ round(src_x)
    vshr.u8     index_uv, #1                @ floor(.../2)
    @ Advance src_xv_[0-3] for the next iteration.
    vadd.u32    src_xv_0, step
    vadd.u32    src_xv_1, step
    vadd.u32    src_xv_2, step
    vadd.u32    src_xv_3, step
    @ Subtract the base offset from the index vectors.
    vdup.u8     offset_uv, index_uv_lo[0]
    vdup.u8     offset, index_lo[0]
    vsub.u8     index_uv, offset_uv
    vsub.u8     index, offset
    @ When we divided index_uv by 2, we put the value into 0.7 format, and
    @ introduced the possibility of an overflow error for the cases where
    @ index_uv_lo[0] is close the maximum (127). For example, .u8(2-126) = 132,
    @ so we need to calculate modulo-128 manually.
    @ Another possibility is to shift in a higher bit from sample_[0-1], but we
    @ can only index 16 values anyway, so a simple mask is sufficient.
    @ Unfortunately, we can't do this with a single vbic instruction.
    vbic.u16    index_uv, #0x0080
    vbic.u16    index_uv, #0x8000

    @ Phase: Load

    @ The nearest Cb/U and Cr/V offset is given as follows:
    @   floor(round(src_x)/2)
    movs        scr3, src_x, LSR #16
    adc         scr3, #0    @ round(src_x), 32.0 format.
    @ Load source values and calculate bilinear filtering co-efficients.
    @ We need to load 32x2 source pixels.
    add         scr0, y, src_x, LSR #16         @ Top Y' block address.
    add         scr1, scr0, rstride_sd, LSR #16 @ Bottom Y' block address.
    add         scr2, u, scr3, LSR #1           @ Cb/U address.
    add         scr3, v, scr3, LSR #1           @ Cr/V address.
    @ Load top Y' data.
    vld1.u8     {data0,data1,data2,data3}, [scr0]
    @ Increment the index vector to get the right-pixel offsets.
    vmov.u8     index2, #1
    vadd.u8     index2, index
    @ Top samples.
    vtbl.u8     src_top_l_lo, {data0,data1,data2,data3}, index_lo
    vtbl.u8     src_top_l_hi, {data0,data1,data2,data3}, index_hi
    vtbl.u8     src_top_r_lo, {data0,data1,data2,data3}, index2_lo
    vtbl.u8     src_top_r_hi, {data0,data1,data2,data3}, index2_hi
    @ Load bottom Y' data.
    vld1.u8     {data0,data1,data2,data3}, [scr1]
    @ Bottom samples.
    vtbl.u8     src_bot_l_lo, {data0,data1,data2,data3}, index_lo
    vtbl.u8     src_bot_l_hi, {data0,data1,data2,data3}, index_hi
    vtbl.u8     src_bot_r_lo, {data0,data1,data2,data3}, index2_lo
    vtbl.u8     src_bot_r_hi, {data0,data1,data2,data3}, index2_hi
    @ Load Cb/U and Cr/V data.
    vld1.u8     {data0,data1}, [scr2] @ Cb / U
    vld1.u8     {data2,data3}, [scr3] @ Cr / V
    @ Advance src_x by 16 source pixels.
    add         src_x, src_dx, LSL #4
    @ Load Cb/U and Cr/V for nearest-pixel filtering.
    pld         [scr0, #32] @ Preload the next Y' top block.
    vtbl.u8     u_filt_lo, {data0,data1}, index_uv_lo
    pld         [scr1, #32] @ Preload the next Y' bottom block.
    vtbl.u8     u_filt_hi, {data0,data1}, index_uv_hi
    pld         [scr2, #16] @ Preload the next Cb/U block.
    vtbl.u8     v_filt_lo, {data2,data3}, index_uv_lo
    pld         [scr3, #16] @ Preload the next Cr/V block.
    vtbl.u8     v_filt_hi, {data2,data3}, index_uv_hi
    
    @ Phase: Filter

    @ Derive w_col_left.
    vmov.u8     w_col_left,  #(1<<7)
    vsub.u8     w_col_left, w_col_right
    @ Calculate right-side contributions.
    vmull.u8    filt_bot_lo,  src_bot_r_lo,   w_col_right_lo
    vmull.u8    filt_bot_hi,  src_bot_r_hi,   w_col_right_hi
    vmull.u8    filt_top_lo,  src_top_r_lo,   w_col_right_lo
    vmull.u8    filt_top_hi,  src_top_r_hi,   w_col_right_hi
    @ Accumulate left-side contributions.
    vmlal.u8    filt_bot_lo,  src_bot_l_lo,   w_col_left_lo
    vmlal.u8    filt_bot_hi,  src_bot_l_hi,   w_col_left_hi
    vmlal.u8    filt_top_lo,  src_top_l_lo,   w_col_left_lo
    vmlal.u8    filt_top_hi,  src_top_l_hi,   w_col_left_hi
    @ Shift down to the 0-255 range.
    vrshr.u16   filt_bot_lo,  #7
    vrshr.u16   filt_bot_hi,  #7
    vrshr.u16   filt_top_lo,  #7
    vrshr.u16   filt_top_hi,  #7

    vdup.u8     coeff_y,     coeff_packed[0]    @ Unpack in advance.

    @ Calculate bottom-row contributions.
    vmul.u16    filt_lo,    filt_bot_lo,  w_row_packed[0]
    vmul.u16    filt_hi,    filt_bot_hi,  w_row_packed[0]
    @ Accumulate top-row contributions.
    vmla.u16    filt_lo,    filt_top_lo,  w_row_packed[1]
    vmla.u16    filt_hi,    filt_top_hi,  w_row_packed[1]
    @ Shift, round and narrow to produce the result.
    vrshrn.u16  y_filt_lo,  filt_lo,  #7
    vrshrn.u16  y_filt_hi,  filt_hi,  #7

    @ Phase: RGB-565

    @ Convert y_filt, u_filt and v_filt to RGB.
    vdup.u8     coeff_u_grn, coeff_packed[3]    @ Unpack in advance.
    @  - Calculate the common Y contribution for each channel.
    vmull.u8    blu16_lo, y_filt_lo, coeff_y
    vmull.u8    blu16_hi, y_filt_hi, coeff_y
    vdup.u8     coeff_u_blu, coeff_packed[4]    @ Unpack in advance.
    @  - Initialize staging areas.
    vmov        red16_lo, blu16_lo
    vmov        red16_hi, blu16_hi
    vmov        grn16_lo, blu16_lo
    vmov        grn16_hi, blu16_hi
    vdup.u8     coeff_v_red, coeff_packed[1]    @ Unpack in advance.
    @  - Negative U contributions for G.
    vmlsl.u8    grn16_lo, u_filt_lo, coeff_u_grn
    vmlsl.u8    grn16_hi, u_filt_hi, coeff_u_grn
    vdup.u8     coeff_v_grn, coeff_packed[2]    @ Unpack in advance.
    @  - Positive U contributions for B.
    vmlal.u8    blu16_lo, u_filt_lo, coeff_u_blu
    vmlal.u8    blu16_hi, u_filt_hi, coeff_u_blu
    vdup.s16    bias_g, bias_packed[1]  @ Unpack G bias in advance.
    @  - Positive V contributions for R.
    vmlal.u8    red16_lo, v_filt_lo, coeff_v_red
    vmlal.u8    red16_hi, v_filt_hi, coeff_v_red
    vdup.s16    bias_r, bias_packed[2]  @ Unpack R bias in advance.
    @  - Negative V contributions for G.
    vmlsl.u8    grn16_lo, v_filt_lo, coeff_v_grn
    vmlsl.u8    grn16_hi, v_filt_hi, coeff_v_grn
    vdup.s16    bias_b, bias_packed[0]  @ Unpack B bias in advance.
    @  - Add bias.
    vsub.s16    grn16_lo, bias_g
    vsub.s16    grn16_hi, bias_g
    vsub.s16    red16_lo, bias_r
    vsub.s16    red16_hi, bias_r
    vsub.s16    blu16_lo, bias_b
    vsub.s16    blu16_hi, bias_b

    @  - Scale back to 8 bits.
    vqrshrun.s16    grn_lo, grn16_lo, #coeff_divisor_imm
    vqrshrun.s16    grn_hi, grn16_hi, #coeff_divisor_imm
    vqrshrun.s16    red_lo, red16_lo, #coeff_divisor_imm
    vqrshrun.s16    red_hi, red16_hi, #coeff_divisor_imm
    vqrshrun.s16    blu_lo, blu16_lo, #coeff_divisor_imm
    vqrshrun.s16    blu_hi, blu16_hi, #coeff_divisor_imm
    
    @  - Convert to RGB-565.
    vshll.u8        rgb565_lo,  red_lo,     #8
    vshll.u8        rgb565_hi,  red_hi,     #8
    vshll.u8        grn16_lo,   grn_lo,     #8
    vshll.u8        grn16_hi,   grn_hi,     #8
    vshll.u8        blu16_lo,   blu_lo,     #8
    vshll.u8        blu16_hi,   blu_hi,     #8
    vsri.u16        rgb565_lo,  grn16_lo,   #5
    vsri.u16        rgb565_hi,  grn16_hi,   #5
    vsri.u16        rgb565_lo,  blu16_lo,   #11
    vsri.u16        rgb565_hi,  blu16_hi,   #11

    @ Phase: Store

    @ Store the ARGB result.
    vst1.u16 {rgb565_lo_lo,rgb565_lo_hi,rgb565_hi_lo,rgb565_hi_hi}, [dst]!

    @ Phase: Column Iteration

    @ End the column loop.
    @ There are three conditions to test:
    @ Let d = dst_w - dst_x.
    @ d >= 32:          The next block fits wholly within the destination image.
    @ d < 32 && d > 16: The next block fits partially within the destination image.
    @ d <= 16:          The next block does not fit in the destination image.
    
    @ Common case: The next block fits wholly within the destination image.
    mov     scr0, dst_wh, LSR #16
    sub     scr0, dst_xy, LSR #16   @ Find 'd'.
    cmp     scr0, #32
    @ Advance dst_xy (in 16:16 format) to the next block.
    add     dst_xy, #(16<<16)

    bhs     2b  @ Loop if d >= 23.
    
    cmp     scr0, #16
    ble     3f  @ Exit column loop if d <= 16.

    @ The next block overlaps the destination image border. We apply offsets
    @ to various loop counters and pointers in order to compensate. This means
    @ that we may re-calculate some data values, but the overhead of this should
    @ be low, and it's probably quicker than running a scalar implementation in
    @ a loop.
    rsb         scr0,   #32             @ Find the distance to the border (dst).
    mul         scr1,   scr0, src_dx    @ Find the distance to the border (src).
    @ Adjust the destination pointers and offsets.
    sub         dst,    scr0, LSL #1
    sub         dst_xy, scr0, LSL #16
    @ Adjust the source offsets.
    vdup.u32    src_xv_step,  scr1
    sub         src_x,  scr1
    vsub.u32    src_xv_0, src_xv_step
    vsub.u32    src_xv_1, src_xv_step
    vsub.u32    src_xv_2, src_xv_step
    vsub.u32    src_xv_3, src_xv_step
    b           2b  @ Loop.

3:  @ ==== End of column loop. ====
    
    @ Phase: Row Iteration

    @ Reset the source and destination pointers.
    ldm         sp, {y,u,v,dst}

    @ Advance dst_xy to the next row.
    bfc         dst_xy, #16, #16
    add         dst_xy, #1

    @ Check that we're still within the destination image.
    @ Note that dst_y has already been incremented for the next row.
    @ Let d = dst_h - dst_y
    @ d >  0: The next destination row is within the destination image.
    @ d <= 0: The next destination row is outside the destination image.
    mov         scr1, dst_wh, LSL #16
    cmp         scr1, dst_xy, LSL #16
    ble         4f  @ Exit the column loop if d <= 0.

    @ Calculate src_y.
    mov         scr0, dst_xy, LSL #16       @ Get dst_y in 16.16 format.
    umull       scr0, scr1, src_dy, scr0    @ Result is in 32.32 format.
    @  - If we're down-scaling (and thus src_dy >= 1), add the initial offset of
    @    (src_dy-1)/2 to src_y.
    subs        scr2, src_dy, #(1<<16)
    movmi       scr2, #0
    adds        scr0, scr2, LSL #15 @ Divide by 2 and convert to 0.32, then add.
    adc         scr1, scr2, LSR #17 @ Divide by 2 and convert to 32.0, then add.
    @ scr0 = src_y, fractional part.
    @ scr1 = src_y, integral part.

    @ y += src_y * src_rstride
    ubfx        scr2, rstride_sd, #16, #16
    mla         y, scr1, scr2, y
    @ u += (floor(round(src_y)/2) * floor(src_rstride/2))
    add         scr1, scr0, LSR #31 @ round(src_y)
    mov         scr1, scr1, LSR #1  @ floor(round(src_y)/2)
    mov         scr2, scr2, LSR #1  @ floor(src_rstride/2)
    mla         u, scr2, scr1, u
    @ v += (floor(round(src_y)/2) * floor(src_rstride/2))
    mla         v, scr2, scr1, v
    @ dst += (dst_y * dst_rstride)
    ubfx        scr1, dst_xy, #0, #16       @ Find dst_y.
    mov         scr2, rstride_sd, LSL #1    @ Find dst_rstride (* 2 bytes per pixel).
    bfc         scr2, #16, #15
    mla         dst, scr1, scr2, dst

    @ Calculate the row weighting vectors in 9.7 format.
    ubfx        scr0, scr0, #25, #7
    rsb         scr1, scr0, #0x0080     @ Top row weighting.
    vmov.u16    w_row_packed[0], scr0
    vmov.u16    w_row_packed[1], scr1

    @ If ((dst_h-dst_y)*src_dy) < 0:
    @   The next two source rows are partially within the source image.
    @ else:
    @   The next two source rows are wholly within the source image.
    mov         scr1, dst_wh, LSL #16
    sub         scr1, dst_xy, LSL #16
    mov         scr1, scr1, ASR #16
    mul         scr1, src_dy, scr1      @ scr = (dst_h-dst_y)*src_dy, 16.16 format.
    cmp         scr1, #0
    bge         1b  @ The next two source rows are wholly within the source image.

    bfc         rstride_sd, #16, #16    @ Make src_rstride 0 to prevent over-read.
    stm         sp, {y,u,v}
    b           1b
    
4:  @ ==== End of row loop. ====

9:  @ We're either finished, or we've given up.
    
    add     sp, sp, #(4*4)

    vpop    {q4-q7}
    pop     {r4-r12, pc}

    .fnend
