/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is a test framework for NEON YUV-to-ARGB code.
 *
 * The Initial Developer of the Original Code is ARM Ltd.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Jacob Bramley <Jacob.Bramley@arm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef RUNNER_H
#define RUNNER_H

#include <stdint.h>
#include "framework.h"

int run(options_t * options);

typedef struct {
    double          time;
    char const *    file_name_tag;  // The tag for the file name, e.g. "bilinear-c".
    char const *    results_tag;    // The tag for the output table, e.g. "[Bilinear][C]".
    scale_rgb565_t  func;
} scale_test_t;

// The list of tests to run.
static scale_test_t tests[] = {

// This is Timothy Terriberry's implementation, modified very slightly to fit
// into this framework.
    {0.0, "bilinear-neon-derf",                     "[Bilinear][derf]",                     ScaleYCbCrToRGB565_Bilinear_NEON_derf},

// The C version is not optimized, but produces a reference image for validity
// checking.
//  {0.0, "bilinear-c",                             "[Bilinear][C]",                        ScaleYCbCrToRGB565_Bilinear_C},

// The scalar load version was my first implementation. It does not over-read
// the source, but suffers a performance penalty as it has to run several
// scalar loads.
//  {0.0, "bilinear-neon-jbramley-scalarload",      "[Bilinear][jbramley][scalar]",         ScaleYCbCrToRGB565_Bilinear_NEON_jbramley},

// The scalar-load-loop version is like the first implementation, but rolls-up
// the load loop to reduce code size. The performance hit is significant!
//  {0.0, "bilinear-neon-jbramley-scalarloadloop",  "[Bilinear][jbramley][scalar][loop]",   ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_loadloop},

// The vtbl version is my most recent, and fastest scaler.
    {0.0, "bilinear-neon-jbramley-vtbl",            "[Bilinear][jbramley][vtbl]",           ScaleYCbCrToRGB565_Bilinear_NEON_jbramley_vtbl},

// This is an empty test, used only to check the overheads of the timing
// mechanism. It produces no output.
//  {0.0, NULL,                                     NULL,                                   ScaleYCbCrToRGB565_Bilinear_NULL},
};

static uint32_t const test_count = sizeof(tests)/sizeof(scale_test_t);

// The following test is taken as the base method, for performance comparison.
// It will therefore always have 100% relative performance.
static scale_test_t * const test_base = &tests[0];

#endif

